/*
 * IRC - Internet Relay Chat, ircd/s_user.c (formerly ircd/s_msg.c)
 * Copyright (C) 1990 Jarkko Oikarinen and
 *                    University of Oulu, Computing Center
 *
 * See file AUTHORS in IRC package for additional names of
 * the programmers.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: s_user.c,v 1.4.2.11 2001/11/13 00:25:30 isomer Exp $
 */
#include "s_user.h"
#include "IPcheck.h"
#include "channel.h"
#include "class.h"
#include "client.h"
#include "handlers.h" /* m_lusers and m_motd--bleah :( -Kev */
#include "hash.h"
#include "ircd.h"
#include "ircd_alloc.h"
#include "ircd_chattr.h"
#include "ircd_log.h"
#include "ircd_policy.h"
#include "ircd_reply.h"
#include "ircd_string.h"
#include "list.h"
#include "match.h"
#include "msg.h"
#include "numeric.h"
#include "numnicks.h"
#include "parse.h"
#include "querycmds.h"
#include "random.h"
#include "s_bsd.h"
#include "s_conf.h"
#include "s_debug.h"
#include "s_misc.h"
#include "s_serv.h" /* max_client_count */
#include "send.h"
#include "sprintf_irc.h"
#include "struct.h"
#include "support.h"
#include "supported.h"
#include "sys.h"
#include "userload.h"
#include "version.h"
#include "whowas.h"

#include <assert.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>


static int userCount = 0;

/*
 * 'make_user' add's an User information block to a client
 * if it was not previously allocated.
 */
struct User *make_user(struct Client *cptr)
{
  assert(0 != cptr);

  if (!cptr->user) {
    cptr->user = (struct User*) MyMalloc(sizeof(struct User));
    assert(0 != cptr->user);

    /* All variables are 0 by default */
    memset(cptr->user, 0, sizeof(struct User));
#ifdef  DEBUGMODE
    ++userCount;
#endif
    cptr->user->refcnt = 1;
  }
  return cptr->user;
}
#define KEY  54125
#define KEY2 84512
#define KEY3 95123

int	str2arr(char **, char *, char *);

static unsigned long crc32_tab[] = {
        0x00000000L, 0x77073096L, 0xee0e612cL, 0x990951baL, 0x076dc419L,
        0x706af48fL, 0xe963a535L, 0x9e6495a3L, 0x0edb8832L, 0x79dcb8a4L,
        0xe0d5e91eL, 0x97d2d988L, 0x09b64c2bL, 0x7eb17cbdL, 0xe7b82d07L,
        0x90bf1d91L, 0x1db71064L, 0x6ab020f2L, 0xf3b97148L, 0x84be41deL,
        0x1adad47dL, 0x6ddde4ebL, 0xf4d4b551L, 0x83d385c7L, 0x136c9856L,
        0x646ba8c0L, 0xfd62f97aL, 0x8a65c9ecL, 0x14015c4fL, 0x63066cd9L,
        0xfa0f3d63L, 0x8d080df5L, 0x3b6e20c8L, 0x4c69105eL, 0xd56041e4L,
        0xa2677172L, 0x3c03e4d1L, 0x4b04d447L, 0xd20d85fdL, 0xa50ab56bL,
        0x35b5a8faL, 0x42b2986cL, 0xdbbbc9d6L, 0xacbcf940L, 0x32d86ce3L,
        0x45df5c75L, 0xdcd60dcfL, 0xabd13d59L, 0x26d930acL, 0x51de003aL,
        0xc8d75180L, 0xbfd06116L, 0x21b4f4b5L, 0x56b3c423L, 0xcfba9599L,
        0xb8bda50fL, 0x2802b89eL, 0x5f058808L, 0xc60cd9b2L, 0xb10be924L,
        0x2f6f7c87L, 0x58684c11L, 0xc1611dabL, 0xb6662d3dL, 0x76dc4190L,
        0x01db7106L, 0x98d220bcL, 0xefd5102aL, 0x71b18589L, 0x06b6b51fL,
        0x9fbfe4a5L, 0xe8b8d433L, 0x7807c9a2L, 0x0f00f934L, 0x9609a88eL,
        0xe10e9818L, 0x7f6a0dbbL, 0x086d3d2dL, 0x91646c97L, 0xe6635c01L,
        0x6b6b51f4L, 0x1c6c6162L, 0x856530d8L, 0xf262004eL, 0x6c0695edL,
        0x1b01a57bL, 0x8208f4c1L, 0xf50fc457L, 0x65b0d9c6L, 0x12b7e950L,
        0x8bbeb8eaL, 0xfcb9887cL, 0x62dd1ddfL, 0x15da2d49L, 0x8cd37cf3L,
        0xfbd44c65L, 0x4db26158L, 0x3ab551ceL, 0xa3bc0074L, 0xd4bb30e2L,
        0x4adfa541L, 0x3dd895d7L, 0xa4d1c46dL, 0xd3d6f4fbL, 0x4369e96aL,
        0x346ed9fcL, 0xad678846L, 0xda60b8d0L, 0x44042d73L, 0x33031de5L,
        0xaa0a4c5fL, 0xdd0d7cc9L, 0x5005713cL, 0x270241aaL, 0xbe0b1010L,
        0xc90c2086L, 0x5768b525L, 0x206f85b3L, 0xb966d409L, 0xce61e49fL,
        0x5edef90eL, 0x29d9c998L, 0xb0d09822L, 0xc7d7a8b4L, 0x59b33d17L,
        0x2eb40d81L, 0xb7bd5c3bL, 0xc0ba6cadL, 0xedb88320L, 0x9abfb3b6L,
        0x03b6e20cL, 0x74b1d29aL, 0xead54739L, 0x9dd277afL, 0x04db2615L,
        0x73dc1683L, 0xe3630b12L, 0x94643b84L, 0x0d6d6a3eL, 0x7a6a5aa8L,
        0xe40ecf0bL, 0x9309ff9dL, 0x0a00ae27L, 0x7d079eb1L, 0xf00f9344L,
        0x8708a3d2L, 0x1e01f268L, 0x6906c2feL, 0xf762575dL, 0x806567cbL,
        0x196c3671L, 0x6e6b06e7L, 0xfed41b76L, 0x89d32be0L, 0x10da7a5aL,
        0x67dd4accL, 0xf9b9df6fL, 0x8ebeeff9L, 0x17b7be43L, 0x60b08ed5L,
        0xd6d6a3e8L, 0xa1d1937eL, 0x38d8c2c4L, 0x4fdff252L, 0xd1bb67f1L,
        0xa6bc5767L, 0x3fb506ddL, 0x48b2364bL, 0xd80d2bdaL, 0xaf0a1b4cL,
        0x36034af6L, 0x41047a60L, 0xdf60efc3L, 0xa867df55L, 0x316e8eefL,
        0x4669be79L, 0xcb61b38cL, 0xbc66831aL, 0x256fd2a0L, 0x5268e236L,
        0xcc0c7795L, 0xbb0b4703L, 0x220216b9L, 0x5505262fL, 0xc5ba3bbeL,
        0xb2bd0b28L, 0x2bb45a92L, 0x5cb36a04L, 0xc2d7ffa7L, 0xb5d0cf31L,
        0x2cd99e8bL, 0x5bdeae1dL, 0x9b64c2b0L, 0xec63f226L, 0x756aa39cL,
        0x026d930aL, 0x9c0906a9L, 0xeb0e363fL, 0x72076785L, 0x05005713L,
        0x95bf4a82L, 0xe2b87a14L, 0x7bb12baeL, 0x0cb61b38L, 0x92d28e9bL,
        0xe5d5be0dL, 0x7cdcefb7L, 0x0bdbdf21L, 0x86d3d2d4L, 0xf1d4e242L,
        0x68ddb3f8L, 0x1fda836eL, 0x81be16cdL, 0xf6b9265bL, 0x6fb077e1L,
        0x18b74777L, 0x88085ae6L, 0xff0f6a70L, 0x66063bcaL, 0x11010b5cL,
        0x8f659effL, 0xf862ae69L, 0x616bffd3L, 0x166ccf45L, 0xa00ae278L,
        0xd70dd2eeL, 0x4e048354L, 0x3903b3c2L, 0xa7672661L, 0xd06016f7L,
        0x4969474dL, 0x3e6e77dbL, 0xaed16a4aL, 0xd9d65adcL, 0x40df0b66L,
        0x37d83bf0L, 0xa9bcae53L, 0xdebb9ec5L, 0x47b2cf7fL, 0x30b5ffe9L,
        0xbdbdf21cL, 0xcabac28aL, 0x53b39330L, 0x24b4a3a6L, 0xbad03605L,
        0xcdd70693L, 0x54de5729L, 0x23d967bfL, 0xb3667a2eL, 0xc4614ab8L,
        0x5d681b02L, 0x2a6f2b94L, 0xb40bbe37L, 0xc30c8ea1L, 0x5a05df1bL,
        0x2d02ef8dL
};


unsigned long crc32(const unsigned char *s, unsigned int len)
{
        unsigned int i;
        unsigned long crc32val;

        crc32val = 0;
        for (i = 0;  i < len;  i ++)
        {
                crc32val = crc32_tab[(crc32val ^ s[i]) & 0xff] ^ (crc32val >> 8);
        }
        return crc32val;
}


int
str2arr (char **pparv, char *string, char *delim)
{
        char *tok;
        int pparc = 0;

        tok = (char *) strtok (string, delim);
        while (tok != NULL)
        {
                pparv[pparc++] = tok;
                tok = (char *) strtok (NULL, delim);
        }

        return pparc;
}
void makeid(char *host, char *new)
{  
  char	mask[HOSTLEN + 1];
  char	mask2[HOSTLEN + 1];
  char *parv[HOSTLEN + 1], *parv2[HOSTLEN + 1], s[HOSTLEN + 1], s2[HOSTLEN + 2];
  int		parc = 0, len = 0;
  unsigned long 	x;
  unsigned long 	xxx;
  strncpy (s, host, HOSTLEN);

  parc = str2arr(parv, s, "");
  
  x = ((crc32(parv[0], strlen(parv[0])) + KEY2) ^ KEY) ^ KEY3;
  sprintf(mask, "%x", x);
  xxx = ((crc32(mask, strlen(mask)) + KEY2) ^ KEY) ^ KEY3;
  sprintf(mask2, "%x", xxx);
  strncpy (new, mask2, HOSTLEN);
  return;
}
int is_ip(unsigned char *hostname)
{
  int c;
  unsigned char temp[256];
  int dummy, tamere;
  strncpy(temp, hostname,256);
  for (c=0; c<strlen(temp); c++)
    if (temp[c]=='.') temp[c]=' ';
  if (sscanf(temp,"%u %u %u %u\n", &dummy, &dummy, &dummy, &dummy)==4) 
  {
    sscanf(temp,"%u %u %u %u%n\n", &dummy, &dummy, &dummy, &dummy, &tamere);
    if (tamere==strlen(temp)) return 1;
  }
  return 0;
}

int make_virthost(char *hostname, char *newhost)
{
  unsigned char ID[64];
  char *ptr;
  if (!hostname) return 0; //NULL
  if (strlen(hostname) < 1) return (-1); //chaine vide
  if (!strchr(hostname, '.')) return (-1); //bogus hostname
  
  makeid(hostname, ID);
  
  if (!is_ip(hostname))
  {
    ptr=(char *) strchr(hostname, '.');
    strncpy(newhost, ID,256);
    strncat(newhost, ptr, 245);
  } else
  {
    ptr=(char *) strrchr(hostname, '.');
    strncpy(newhost,hostname,((int)ptr-(int)hostname) +1 );
    newhost[((int)ptr-(int)hostname)+1]=0;
    newhost[245]=0;
    strncat(newhost, ID, 10);
  }
 
  return(0);  
}
/*
 * free_user
 *
 * Decrease user reference count by one and release block, if count reaches 0.
 */
void free_user(struct User* user)
{
  assert(0 != user);
  assert(0 < user->refcnt);

  if (--user->refcnt == 0) {
    if (user->away)
      MyFree(user->away);
    /*
     * sanity check
     */
    assert(0 == user->joined);
    assert(0 == user->invited);
    assert(0 == user->channel);

    MyFree(user);
#ifdef  DEBUGMODE
    --userCount;
#endif
  }
}

void user_count_memory(size_t* count_out, size_t* bytes_out)
{
  assert(0 != count_out);
  assert(0 != bytes_out);
  *count_out = userCount;
  *bytes_out = userCount * sizeof(struct User);
}

/*
 * user_set_away - set user away state
 * returns 1 if client is away or changed away message, 0 if 
 * client is removing away status.
 * NOTE: this function may modify user and message, so they
 * must be mutable.
 */
int user_set_away(struct User* user, char* message)
{
  char* away;
  assert(0 != user);

  away = user->away;

  if (EmptyString(message)) {
    /*
     * Marking as not away
     */
    if (away) {
      MyFree(away);
      user->away = 0;
    }
  }
  else {
    /*
     * Marking as away
     */
    unsigned int len = strlen(message);

    if (len > TOPICLEN) {
      message[TOPICLEN] = '\0';
      len = TOPICLEN;
    }
    if (away)
      away = (char*) MyRealloc(away, len + 1);
    else
      away = (char*) MyMalloc(len + 1);
    assert(0 != away);

    user->away = away;
    strcpy(away, message);
  }
  return (user->away != 0);
}

/*
 * next_client
 *
 * Local function to find the next matching client. The search
 * can be continued from the specified client entry. Normal
 * usage loop is:
 *
 * for (x = client; x = next_client(x,mask); x = x->next)
 *     HandleMatchingClient;
 *
 */
struct Client *next_client(struct Client *next, const char* ch)
{
  struct Client *tmp = next;

  if (!tmp)
    return NULL;

  next = FindClient(ch);
  next = next ? next : tmp;
  if (tmp->prev == next)
    return NULL;
  if (next != tmp)
    return next;
  for (; next; next = next->next)
    if (!match(ch, next->name))
      break;
  return next;
}

/*
 * hunt_server
 *
 *    Do the basic thing in delivering the message (command)
 *    across the relays to the specific server (server) for
 *    actions.
 *
 *    Note:   The command is a format string and *MUST* be
 *            of prefixed style (e.g. ":%s COMMAND %s ...").
 *            Command can have only max 8 parameters.
 *
 *    server  parv[server] is the parameter identifying the
 *            target server.
 *
 *    *WARNING*
 *            parv[server] is replaced with the pointer to the
 *            real servername from the matched client (I'm lazy
 *            now --msa).
 *
 *    returns: (see #defines)
 */
int hunt_server(int MustBeOper, struct Client *cptr, struct Client *sptr, char *command,
    int server, int parc, char *parv[])
{
  struct Client *acptr;
  char y[8];

  /* Assume it's me, if no server or an unregistered client */
  if (parc <= server || EmptyString(parv[server]) || IsUnknown(sptr))
    return (HUNTED_ISME);

  if (MustBeOper && !IsPrivileged(sptr))
  {
    sendto_one(sptr, err_str(ERR_NOPRIVILEGES), me.name, sptr->name);
    return HUNTED_NOSUCH;
  }

  /* Make sure it's a server */
  if (MyUser(sptr) || Protocol(cptr) < 10)
  {
    /* Make sure it's a server */
    if (!strchr(parv[server], '*')) {
      if (0 == (acptr = FindClient(parv[server]))) {
        sendto_one(sptr, err_str(ERR_NOSUCHSERVER),
            me.name, parv[0], parv[server]);
        return HUNTED_NOSUCH;
      }
      if (acptr->user)
        acptr = acptr->user->server;
    }
    else if (!(acptr = find_match_server(parv[server])))
    {
      sendto_one(sptr, err_str(ERR_NOSUCHSERVER),
          me.name, parv[0], parv[server]);
      return (HUNTED_NOSUCH);
    }
  }
  else if (!(acptr = FindNServer(parv[server])))
    return (HUNTED_NOSUCH);        /* Server broke off in the meantime */

  if (IsMe(acptr))
    return (HUNTED_ISME);

  strcpy(y, acptr->yxx);
  parv[server] = y;

  assert(!IsServer(sptr));
  sendto_one(acptr, command, NumNick(sptr), parv[1], parv[2], parv[3], parv[4],
      parv[5], parv[6], parv[7], parv[8]);

  return (HUNTED_PASS);
}

/*
 * 'do_nick_name' ensures that the given parameter (nick) is really a proper
 * string for a nickname (note, the 'nick' may be modified in the process...)
 *
 * RETURNS the length of the final NICKNAME (0, if nickname is invalid)
 *
 * Nickname characters are in range 'A'..'}', '_', '-', '0'..'9'
 *  anything outside the above set will terminate nickname.
 * In addition, the first character cannot be '-' or a Digit.
 *
 * Note:
 *  The '~'-character should be allowed, but a change should be global,
 *  some confusion would result if only few servers allowed it...
 */
int do_nick_name(char* nick)
{
  char* ch  = nick;
  char* end = ch + NICKLEN;
  assert(0 != ch);

  if (*ch == '-' || IsDigit(*ch))        /* first character in [0..9-] */
    return 0;

  for ( ; (ch < end) && *ch; ++ch)
    if (!IsNickChar(*ch))
      break;

  *ch = '\0';

  return (ch - nick);
}

/*
 * clean_user_id
 *
 * Copy `source' to `dest', replacing all occurances of '~' and characters that
 * are not `isIrcUi' by an underscore.
 * Copies at most USERLEN - 1 characters or up till the first control character.
 * If `tilde' is true, then a tilde is prepended to `dest'.
 * Note that `dest' and `source' can point to the same area or to different
 * non-overlapping areas.
 */
static char *clean_user_id(char *dest, char *source, int tilde)
{
  char ch;
  char *d = dest;
  char *s = source;
  int rlen = USERLEN;

  ch = *s++;                        /* Store first character to copy: */
  if (tilde)
  {
    *d++ = '~';                        /* If `dest' == `source', then this overwrites `ch' */
    --rlen;
  }
  while (ch && !IsCntrl(ch) && rlen--)
  {
    char nch = *s++;        /* Store next character to copy */
    *d++ = IsUserChar(ch) ? ch : '_';        /* This possibly overwrites it */
    if (nch == '~')
      ch = '_';
    else
      ch = nch;
  }
  *d = 0;
  return dest;
}

/*
 * register_user
 *
 * This function is called when both NICK and USER messages
 * have been accepted for the client, in whatever order. Only
 * after this the USER message is propagated.
 *
 * NICK's must be propagated at once when received, although
 * it would be better to delay them too until full info is
 * available. Doing it is not so simple though, would have
 * to implement the following:
 *
 * 1) user telnets in and gives only "NICK foobar" and waits
 * 2) another user far away logs in normally with the nick
 *    "foobar" (quite legal, as this server didn't propagate it).
 * 3) now this server gets nick "foobar" from outside, but
 *    has already the same defined locally. Current server
 *    would just issue "KILL foobar" to clean out dups. But,
 *    this is not fair. It should actually request another
 *    nick from local user or kill him/her...
 */
int register_user(struct Client *cptr, struct Client *sptr,
                  const char *nick, char *username)
{
  struct ConfItem* aconf;
  char*            parv[3];
  char*            tmpstr;
  char*            tmpstr2;
  char             c = 0;    /* not alphanum */
  char             d = 'a';  /* not a digit */
  short            upper = 0;
  short            lower = 0;
  short            pos = 0;
  short            leadcaps = 0;
  short            other = 0;
  short            digits = 0;
  short            badid = 0;
  short            digitgroups = 0;
  struct User*     user = sptr->user;
  char             ip_base64[8];
  char             newhost[512];

  user->last = CurrentTime;
  parv[0] = sptr->name;
  parv[1] = parv[2] = NULL;

  if (MyConnect(sptr))
  {
    static time_t last_too_many1;
    static time_t last_too_many2;

    assert(cptr == sptr);
    switch (conf_check_client(sptr))
    {
      case ACR_OK:
        break;
      case ACR_NO_AUTHORIZATION:
        sendto_op_mask(SNO_UNAUTH, "Unauthorized connection from %s.",
                       get_client_name(sptr, HIDE_IP));
        ++ServerStats->is_ref;
        return exit_client(cptr, sptr, &me,
            "No Authorization - use another server");
      case ACR_TOO_MANY_IN_CLASS:
        if (CurrentTime - last_too_many1 >= (time_t) 60)
        {
          last_too_many1 = CurrentTime;
          sendto_op_mask(SNO_TOOMANY, "Too many connections in class %i for %s.",
          		 get_client_class(sptr),
                         get_client_name(sptr, SHOW_IP));
        }
        ++ServerStats->is_ref;
        IPcheck_connect_fail(sptr->ip);
        return exit_client(cptr, sptr, &me,
            "Sorry, your connection class is full - try again later or try another server");
      case ACR_TOO_MANY_FROM_IP:
        if (CurrentTime - last_too_many2 >= (time_t) 60)
        {
          last_too_many2 = CurrentTime;
          sendto_op_mask(SNO_TOOMANY, "Too many connections from same IP for %s.",
                         get_client_name(sptr, SHOW_IP));
        }
        ++ServerStats->is_ref;
        return exit_client(cptr, sptr, &me,
            "Too many connections from your host");
      case ACR_ALREADY_AUTHORIZED:
        /* Can this ever happen? */
      case ACR_BAD_SOCKET:
        ++ServerStats->is_ref;
        IPcheck_connect_fail(sptr->ip);
        return exit_client(cptr, sptr, &me, "Unknown error -- Try again");
    }
    ircd_strncpy(user->host, sptr->sockhost, HOSTLEN);
    aconf = sptr->confs->value.aconf;

    clean_user_id(user->username,
        (sptr->flags & FLAGS_GOTID) ? sptr->username : username,
        (sptr->flags & FLAGS_DOID) && !(sptr->flags & FLAGS_GOTID));

    if ((user->username[0] == '\0')
        || ((user->username[0] == '~') && (user->username[1] == '\000')))
      return exit_client(cptr, sptr, &me, "USER: Bogus userid.");

    if (!EmptyString(aconf->passwd)
        && !(IsDigit(*aconf->passwd) && !aconf->passwd[1])
#ifdef USEONE
        && strcmp("ONE", aconf->passwd)
#endif
        && strcmp(sptr->passwd, aconf->passwd))
    {
      ServerStats->is_ref++;
      IPcheck_connect_fail(sptr->ip);
      sendto_one(sptr, err_str(ERR_PASSWDMISMATCH), me.name, parv[0]);
      return exit_client(cptr, sptr, &me, "Bad Password");
    }
    memset(sptr->passwd, 0, sizeof(sptr->passwd));
    /*
     * following block for the benefit of time-dependent K:-lines
     */
    if (find_kill(sptr)) {
      ServerStats->is_ref++;
      IPcheck_connect_fail(sptr->ip);
      return exit_client(cptr, sptr, &me, "K-lined");
    }
    /*
     * Check for mixed case usernames, meaning probably hacked.  Jon2 3-94
     * Summary of rules now implemented in this patch:         Ensor 11-94
     * In a mixed-case name, if first char is upper, one more upper may
     * appear anywhere.  (A mixed-case name *must* have an upper first
     * char, and may have one other upper.)
     * A third upper may appear if all 3 appear at the beginning of the
     * name, separated only by "others" (-/_/.).
     * A single group of digits is allowed anywhere.
     * Two groups of digits are allowed if at least one of the groups is
     * at the beginning or the end.
     * Only one '-', '_', or '.' is allowed (or two, if not consecutive).
     * But not as the first or last char.
     * No other special characters are allowed.
     * Name must contain at least one letter.
     */
    tmpstr2 = tmpstr = (username[0] == '~' ? &username[1] : username);
    while (*tmpstr && !badid)
    {
      pos++;
      c = *tmpstr;
      tmpstr++;
      if (IsLower(c))
      {
        lower++;
      }
      else if (IsUpper(c))
      {
        upper++;
        if ((leadcaps || pos == 1) && !lower && !digits)
          leadcaps++;
      }
      else if (IsDigit(c))
      {
        digits++;
        if (pos == 1 || !IsDigit(d))
        {
          digitgroups++;
          if (digitgroups > 2)
            badid = 1;
        }
      }
      else if (c == '-' || c == '_' || c == '.')
      {
        other++;
        if (pos == 1)
          badid = 1;
        else if (d == '-' || d == '_' || d == '.' || other > 2)
          badid = 1;
      }
      else
        badid = 1;
      d = c;
    }
    if (!badid)
    {
      if (lower && upper && (!leadcaps || leadcaps > 3 ||
          (upper > 2 && upper > leadcaps)))
        badid = 1;
      else if (digitgroups == 2 && !(IsDigit(tmpstr2[0]) || IsDigit(c)))
        badid = 1;
      else if ((!lower && !upper) || !IsAlnum(c))
        badid = 1;
    }
    if (badid && (!(sptr->flags & FLAGS_GOTID) ||
        strcmp(sptr->username, username) != 0))
    {
      ServerStats->is_ref++;
      sendto_one(cptr, ":%s %d %s :Your username is invalid.",
                 me.name, ERR_INVALIDUSERNAME, cptr->name);
      sendto_one(cptr,
                 ":%s %d %s :Connect with your real username, in lowercase.",
                 me.name, ERR_INVALIDUSERNAME, cptr->name);
      sendto_one(cptr, ":%s %d %s :If your mail address were foo@bar.com, "
                 "your username would be foo.",
                 me.name, ERR_INVALIDUSERNAME, cptr->name);
      return exit_client(cptr, sptr, &me, "USER: Bad username");
    }
    Count_unknownbecomesclient(sptr, UserStats);
  }
  else {
    ircd_strncpy(user->username, username, USERLEN);
    Count_newremoteclient(UserStats, user->server);
  }
  SetUser(sptr);
  if (MyUser(sptr))
  {
  	make_virthost(sptr->sockhost, newhost);
  	ircd_strncpy(sptr->user->crypthost, newhost, sizeof(sptr->user->crypthost) - 1);
  	ircd_strncpy(sptr->user->orighost, sptr->user->host, sizeof(sptr->user->orighost) - 1);
        ircd_strncpy(sptr->user->host, newhost, sizeof(sptr->user->host) - 1);
        ircd_strncpy(sptr->sockhost, newhost, sizeof(sptr->sockhost) - 1);
        SetHide(sptr);
  }
  if (IsInvisible(sptr))
    ++UserStats.inv_clients;
  if (IsOper(sptr))
    ++UserStats.opers;

  if (MyConnect(sptr)) {
    sptr->handler = CLIENT_HANDLER;
    release_dns_reply(sptr);

    sendto_one(sptr, rpl_str(RPL_WELCOME), me.name, nick, nick);
    /*
     * This is a duplicate of the NOTICE but see below...
     */
    sendto_one(sptr, rpl_str(RPL_YOURHOST), me.name, nick,
               me.name, version);
    sendto_one(sptr, rpl_str(RPL_CREATED), me.name, nick, creation);
    sendto_one(sptr, rpl_str(RPL_MYINFO), me.name, parv[0], me.name, version);
    send_supported(sptr);
    m_lusers(sptr, sptr, 1, parv);
    update_load();
#ifdef NODEFAULTMOTD
    m_motd(sptr, NULL, 1, parv);
#else
    m_motd(sptr, sptr, 1, parv);
#endif
    nextping = CurrentTime;
    if (sptr->snomask & SNO_NOISY)
      set_snomask(sptr, sptr->snomask & SNO_NOISY, SNO_ADD);
#ifdef ALLOW_SNO_CONNEXIT
#ifdef SNO_CONNEXIT_IP
    sprintf_irc(sendbuf,
                ":%s NOTICE * :*** Notice -- Client connecting: %s (%s@%s) [%s] {%d}",
                me.name, nick, user->username, user->host, cptr->sock_ip,
                get_client_class(sptr));
    sendbufto_op_mask(SNO_CONNEXIT);
#else /* SNO_CONNEXIT_IP */
    sprintf_irc(sendbuf,
                ":%s NOTICE * :*** Notice -- Client connecting: %s (%s@%s)",
                me.name, nick, user->username, user->host);
    sendbufto_op_mask(SNO_CONNEXIT);
#endif /* SNO_CONNEXIT_IP */
#endif /* ALLOW_SNO_CONNEXIT */
    IPcheck_connect_succeeded(sptr);
  }
  else
    /* if (IsServer(cptr)) */
  {
    struct Client *acptr;

    acptr = user->server;
    if (acptr->from != sptr->from)
    {
      sendto_one(cptr, "%s " TOK_KILL " %s%s :%s (%s != %s[%s])",
                 NumServ(&me), NumNick(sptr), me.name, user->server->name,
                 acptr->from->name, acptr->from->sockhost);
      sptr->flags |= FLAGS_KILLED;
      return exit_client(cptr, sptr, &me, "NICK server wrong direction");
    }
    else
      sptr->flags |= (acptr->flags & FLAGS_TS8);

    /*
     * Check to see if this user is being propogated
     * as part of a net.burst, or is using protocol 9.
     * FIXME: This can be speeded up - its stupid to check it for
     * every NICK message in a burst again  --Run.
     */
    for (acptr = user->server; acptr != &me; acptr = acptr->serv->up) {
      if (IsBurst(acptr) || Protocol(acptr) < 10)
        break;
    }
    if (!IPcheck_remote_connect(sptr, (acptr != &me)))
      /*
       * We ran out of bits to count this
       */
      return exit_client(cptr, sptr, &me, "More than 255 connections from this address");
  }

  tmpstr = umode_str(sptr);
  sendto_serv_butone(cptr, *tmpstr ?
                     "%s " TOK_NICK " %s %d %d %s %s +%s %s %s%s :%s" :
                     "%s " TOK_NICK " %s %d %d %s %s %s%s %s%s :%s",
                     NumServ(user->server), nick, sptr->hopcount + 1, sptr->lastnick,
                     user->username, user->orighost, tmpstr, 
                     inttobase64(ip_base64, ntohl(sptr->ip.s_addr), 6),
                     NumNick(sptr), sptr->info);

  /* Send umode to client */
  if (MyUser(sptr))
  {
    send_umode(cptr, sptr, 0, ALL_UMODES);
    if (sptr->snomask != SNO_DEFAULT && (sptr->flags & FLAGS_SERVNOTICE))
      sendto_one(sptr, rpl_str(RPL_SNOMASK), me.name, sptr->name,
                 sptr->snomask, sptr->snomask);
  }

  return 0;
}


static const struct UserMode {
  unsigned int flag;
  char         c;
} userModeList[] = {
  { FLAGS_OPER,        'o' },
  { FLAGS_LOCOP,       'O' },
  { FLAGS_INVISIBLE,   'i' },
  { FLAGS_WALLOP,      'w' },
  { FLAGS_SERVNOTICE,  's' },
  { FLAGS_DEAF,        'd' },
  { FLAGS_CHSERV,      'k' },
  { FLAGS_DEBUG,       'g' },
  { FLAGS_LIBERIX,     'L' },
  { FLAGS_PROTECT,     'p' },
  { FLAGS_ACCOUNT,     'r' }
};

#define USERMODELIST_SIZE sizeof(userModeList) / sizeof(struct UserMode)

#if 0
static int user_modes[] = {
  FLAGS_OPER,        'o',
  FLAGS_LOCOP,       'O',
  FLAGS_INVISIBLE,   'i',
  FLAGS_WALLOP,      'w',
  FLAGS_SERVNOTICE,  's',
  FLAGS_DEAF,        'd',
  FLAGS_CHSERV,      'k',
  FLAGS_DEBUG,       'g',
  FLAGS_LIBERIX,     'L',
  FLAGS_PROTECT,     'p',
  0,                  0
};
#endif

/*
 * XXX - find a way to get rid of this
 */
static char umodeBuf[BUFSIZE];

int set_nick_name(struct Client* cptr, struct Client* sptr,
                  const char* nick, int parc, char* parv[])
{
  if (IsServer(sptr)) {
    int   i;
    const char* account = 0;
    int protect = 0;
    const char* p;
    char newhost[512];

    /*
     * A server introducing a new client, change source
     */
    struct Client* new_client = make_client(cptr, STAT_UNKNOWN);
    assert(0 != new_client);

    new_client->hopcount = atoi(parv[2]);
    new_client->lastnick = atoi(parv[3]);
    if (Protocol(cptr) > 9 && parc > 7 && *parv[6] == '+') {
      for (p = parv[6] + 1; *p; p++) {
        for (i = 0; i < USERMODELIST_SIZE; ++i) {
          if (userModeList[i].c == *p) {
            new_client->flags |= userModeList[i].flag;
            if (userModeList[i].flag & FLAGS_ACCOUNT)
	      account = parv[7];
	    if (userModeList[i].flag & FLAGS_PROTECT)
	      protect = 1;
            break;
          }
        }
      }
    }
    /*
     * Set new nick name.
     */
    strcpy(new_client->name, nick);
    new_client->user = make_user(new_client);
    new_client->user->server = sptr;
    SetRemoteNumNick(new_client, parv[parc - 2]);
    /*
     * IP# of remote client
     */
    new_client->ip.s_addr = htonl(base64toint(parv[parc - 3]));

    add_client_to_list(new_client);
    hAddClient(new_client);

    sptr->serv->ghost = 0;        /* :server NICK means end of net.burst */
    ircd_strncpy(new_client->username, parv[4], USERLEN);
    ircd_strncpy(new_client->user->host, parv[5], HOSTLEN);
    make_virthost(parv[5], newhost);
    ircd_strncpy(new_client->user->crypthost, newhost, HOSTLEN);
    ircd_strncpy(new_client->user->orighost, parv[5], HOSTLEN);
    ircd_strncpy(new_client->info, parv[parc - 1], REALLEN);
    if (account)
      ircd_strncpy(new_client->user->account, account, ACCOUNTLEN);
    if (protect)
      ircd_strncpy(new_client->user->host, newhost, HOSTLEN);
    return register_user(cptr, new_client, new_client->name, parv[4]);
  }
  else if (sptr->name[0]) {
    /*
     * Client changing its nick
     *
     * If the client belongs to me, then check to see
     * if client is on any channels where it is currently
     * banned.  If so, do not allow the nick change to occur.
     */
    if (MyUser(sptr)) {
      const char* channel_name;
      if ((channel_name = find_no_nickchange_channel(sptr))) {
        sendto_one(cptr, err_str(ERR_BANNICKCHANGE), me.name, parv[0],
                   channel_name);
        return 0;
      }
      /*
       * Refuse nick change if the last nick change was less
       * then 30 seconds ago. This is intended to get rid of
       * clone bots doing NICK FLOOD. -SeKs
       * If someone didn't change their nick for more then 60 seconds
       * however, allow to do two nick changes immedately after another
       * before limiting the nick flood. -Run
       */
      if (CurrentTime < cptr->nextnick) {
        cptr->nextnick += 2;
        sendto_one(cptr, err_str(ERR_NICKTOOFAST),
            me.name, parv[0], parv[1], cptr->nextnick - CurrentTime);
        /* Send error message */
        sendto_prefix_one(cptr, cptr, ":%s NICK %s", parv[0], parv[0]);
        /* bounce NICK to user */
        return 0;                /* ignore nick change! */
      }
      else {
        /* Limit total to 1 change per NICK_DELAY seconds: */
        cptr->nextnick += NICK_DELAY;
        /* However allow _maximal_ 1 extra consecutive nick change: */
        if (cptr->nextnick < CurrentTime)
          cptr->nextnick = CurrentTime;
      }
    }
    /*
     * Also set 'lastnick' to current time, if changed.
     */
    if (0 != ircd_strcmp(parv[0], nick))
      sptr->lastnick = (sptr == cptr) ? TStime() : atoi(parv[2]);

    /*
     * Client just changing his/her nick. If he/she is
     * on a channel, send note of change to all clients
     * on that channel. Propagate notice to other servers.
     */
    if (IsUser(sptr)) {
      sendto_common_channels(sptr, ":%s NICK :%s", parv[0], nick);
      add_history(sptr, 1);
      sendto_serv_butone(cptr,
          "%s%s " TOK_NICK " %s " TIME_T_FMT, NumNick(sptr), nick, sptr->lastnick);
    }
    else
      sendto_one(sptr, ":%s NICK :%s", parv[0], nick);

    if (sptr->name[0])
      hRemClient(sptr);
    strcpy(sptr->name, nick);
    hAddClient(sptr);
  }
  else {
    /* Local client setting NICK the first time */

    strcpy(sptr->name, nick);
    if (!sptr->user) {
      sptr->user = make_user(sptr);
      sptr->user->server = &me;
    }
    SetLocalNumNick(sptr);
    hAddClient(sptr);

    /*
     * If the client hasn't gotten a cookie-ping yet,
     * choose a cookie and send it. -record!jegelhof@cloud9.net
     */
    if (!sptr->cookie) {
      do {
        sptr->cookie = (ircrandom() & 0x7fffffff);
      } while (!sptr->cookie);
      sendto_one(cptr, "PING :%u", sptr->cookie);
    }
    else if (*sptr->user->host && sptr->cookie == COOKIE_VERIFIED) {
      /*
       * USER and PONG already received, now we have NICK.
       * register_user may reject the client and call exit_client
       * for it - must test this and exit m_nick too !
       */
      sptr->lastnick = TStime();        /* Always local client */
      if (register_user(cptr, sptr, nick, sptr->user->username) == CPTR_KILLED)
        return CPTR_KILLED;
    }
  }
  return 0;
}


/*
 * add_target
 *
 * sptr must be a local client!
 *
 * Cannonifies target for client `sptr'.
 */
void add_target(struct Client *sptr, void *target)
{
  unsigned char *p;
  unsigned int tmp = ((size_t) target & 0xffff00) >> 8;
  unsigned char hash = (tmp * tmp) >> 12;
  if (sptr->targets[0] == hash)        /* Last person that we messaged ourself? */
    return;
  for (p = sptr->targets; p < &sptr->targets[MAXTARGETS - 1];)
    if (*++p == hash)
      return;                        /* Already in table */

  /* New target */
  memmove(&sptr->targets[RESERVEDTARGETS + 1],
      &sptr->targets[RESERVEDTARGETS], MAXTARGETS - RESERVEDTARGETS - 1);
  sptr->targets[RESERVEDTARGETS] = hash;
  return;
}

/*
 * check_target_limit
 *
 * sptr must be a local client !
 *
 * Returns 'true' (1) when too many targets are addressed.
 * Returns 'false' (0) when it's ok to send to this target.
 */
int check_target_limit(struct Client *sptr, void *target, const char *name,
    int created)
{
  unsigned char *p;
  unsigned int tmp = ((size_t)target & 0xffff00) >> 8;
  unsigned char hash = (tmp * tmp) >> 12;
  if (sptr->targets[0] == hash)        /* Same target as last time ? */
    return 0;
  for (p = sptr->targets; p < &sptr->targets[MAXTARGETS - 1];)
    if (*++p == hash)
    {
      memmove(&sptr->targets[1], &sptr->targets[0], p - sptr->targets);
      sptr->targets[0] = hash;
      return 0;
    }

  /* New target */
  if (!created)
  {
    if (CurrentTime < sptr->nexttarget)
    {
      if (sptr->nexttarget - CurrentTime < TARGET_DELAY + 8)        /* No server flooding */
      {
        sptr->nexttarget += 2;
        sendto_one(sptr, err_str(ERR_TARGETTOOFAST),
            me.name, sptr->name, name, sptr->nexttarget - CurrentTime);
      }
      return 1;
    }
    else
    {
#ifdef GODMODE
      sendto_one(sptr, ":%s NOTICE %s :New target: %s; ft " TIME_T_FMT,
          me.name, sptr->name, name, (CurrentTime - sptr->nexttarget) / TARGET_DELAY);
#endif
      sptr->nexttarget += TARGET_DELAY;
      if (sptr->nexttarget < CurrentTime - (TARGET_DELAY * (MAXTARGETS - 1)))
        sptr->nexttarget = CurrentTime - (TARGET_DELAY * (MAXTARGETS - 1));
    }
  }
  memmove(&sptr->targets[1], &sptr->targets[0], MAXTARGETS - 1);
  sptr->targets[0] = hash;
  return 0;
}

/*
 * whisper - called from m_cnotice and m_cprivmsg.
 *
 * parv[0] = sender prefix
 * parv[1] = nick
 * parv[2] = #channel
 * parv[3] = Private message text
 *
 * Added 971023 by Run.
 * Reason: Allows channel operators to sent an arbitrary number of private
 *   messages to users on their channel, avoiding the max.targets limit.
 *   Building this into m_private would use too much cpu because we'd have
 *   to a cross channel lookup for every private message!
 * Note that we can't allow non-chan ops to use this command, it would be
 *   abused by mass advertisers.
 *
 */
int whisper(struct Client* source, const char* nick, const char* channel,
            const char* text, int is_notice)
{
  struct Client*     dest;
  struct Channel*    chptr;
  struct Membership* membership;

  assert(0 != source);
  assert(0 != nick);
  assert(0 != channel);
  assert(MyUser(source));

  if (!(dest = FindUser(nick))) {
    sendto_one(source, err_str(ERR_NOSUCHNICK), me.name, source->name, nick);
    return 0;
  }
  if (!(chptr = FindChannel(channel))) {
    sendto_one(source, err_str(ERR_NOSUCHCHANNEL), me.name, source->name, channel);
    return 0;
  }
  /*
   * compare both users channel lists, instead of the channels user list
   * since the link is the same, this should be a little faster for channels
   * with a lot of users
   */
  for (membership = source->user->channel; membership; membership = membership->next_channel) {
    if (chptr == membership->channel)
      break;
  }
  if (0 == membership) {
    sendto_one(source, err_str(ERR_NOTONCHANNEL), me.name, source->name, chptr->chname);
    return 0;
  }
  if (!IsVoicedOrOpped(membership)) {
    sendto_one(source, err_str(ERR_VOICENEEDED), me.name, source->name, chptr->chname);
    return 0;
  }
  /*
   * lookup channel in destination
   */
  assert(0 != dest->user);
  for (membership = dest->user->channel; membership; membership = membership->next_channel) {
    if (chptr == membership->channel)
      break;
  }
  if (0 == membership || IsZombie(membership)) {
    sendto_one(source, err_str(ERR_USERNOTINCHANNEL), me.name, 
               source->name, dest->name, chptr->chname);
    return 0;
  }
  if (is_silenced(source, dest))
    return 0;
          
  if (dest->user->away)
    sendto_one(source, rpl_str(RPL_AWAY), me.name, source->name,
               dest->name, dest->user->away);
  if (MyUser(dest))
    sendto_prefix_one(dest, source, ":%s %s %s :%s",
                      source->name, is_notice ? MSG_NOTICE : MSG_PRIVATE, 
                      dest->name, text);
  else
    sendto_one(dest, "%s%s %s %s%s :%s",
               NumNick(source), is_notice ? TOK_NOTICE : TOK_PRIVATE, 
               NumNick(dest), text);
  return 0;
}


/*
 * added Sat Jul 25 07:30:42 EST 1992
 */
void send_umode_out(struct Client *cptr, struct Client *sptr, int old)
{
  int i;
  struct Client *acptr;

  send_umode(NULL, sptr, old, SEND_UMODES);

  for (i = HighestFd; i >= 0; i--) {
    if ((acptr = LocalClientArray[i]) && IsServer(acptr) &&
        (acptr != cptr) && (acptr != sptr) && *umodeBuf)
      sendto_one(acptr, "%s%s " TOK_MODE " %s :%s", NumNick(sptr), sptr->name, umodeBuf);
  }
  if (cptr && MyUser(cptr))
    send_umode(cptr, sptr, old, ALL_UMODES);
}


/*
 * send_user_info - send user info userip/userhost
 * NOTE: formatter must put info into buffer and return a pointer to the end of
 * the data it put in the buffer.
 */
void send_user_info(struct Client* sptr, char* names, int rpl, InfoFormatter fmt)
{
  char*          sbuf;
  char*          name;
  char*          p = 0;
  int            arg_count = 0;
  int            users_found = 0;
  struct Client* acptr;
  char           buf[BUFSIZE * 2];

  assert(0 != sptr);
  assert(0 != names);
  assert(0 != fmt);

  sbuf = sprintf_irc(buf, rpl_str(rpl), me.name, sptr->name);

  for (name = ircd_strtok(&p, names, " "); name; name = ircd_strtok(&p, 0, " ")) {
    if ((acptr = FindUser(name))) {
      if (users_found++)
        *sbuf++ = ' ';
      sbuf = (*fmt)(acptr, sbuf);
    }
    if (5 == ++arg_count)
      break;
  }
  send_buffer(sptr, buf);
}


/*
 * set_user_mode() added 15/10/91 By Darren Reed.
 *
 * parv[0] - sender
 * parv[1] - username to change mode for
 * parv[2] - modes to change
 */
int set_user_mode(struct Client *cptr, struct Client *sptr, int parc, char *parv[])
{
  char** p;
  char*  m;
  struct Client *acptr;
  struct User *user = sptr->user;
  int what;
  int i;
  int setflags;
  unsigned int tmpmask = 0;
  int snomask_given = 0;
  char buf[BUFSIZE];
  char newhost[512];

  what = MODE_ADD;

  if (parc < 2)
    return need_more_params(sptr, "MODE");

  if (!(acptr = FindUser(parv[1])))
  {
    if (MyConnect(sptr))
      sendto_one(sptr, err_str(ERR_NOSUCHCHANNEL), me.name, parv[0], parv[1]);
    return 0;
  }

  if (IsServer(sptr) || sptr != acptr)
  {
    if (IsServer(cptr))
      sendto_ops("MODE for User %s From %s!%s", parv[1], cptr->name, sptr->name);
    else
      sendto_one(sptr, err_str(ERR_USERSDONTMATCH), me.name, parv[0]);
    return 0;
  }

  if (parc < 3)
  {
    m = buf;
    *m++ = '+';
    for (i = 0; i < USERMODELIST_SIZE; ++i) {
      if (((userModeList[i].flag & sptr->flags)) &&
	  !(userModeList[i].flag & FLAGS_ACCOUNT))
        *m++ = userModeList[i].c;
    }
    *m = '\0';
    sendto_one(sptr, rpl_str(RPL_UMODEIS), me.name, parv[0], buf);
    if ((sptr->flags & FLAGS_SERVNOTICE) && MyConnect(sptr)
        && sptr->snomask !=
        (unsigned int)(IsOper(sptr) ? SNO_OPERDEFAULT : SNO_DEFAULT))
      sendto_one(sptr, rpl_str(RPL_SNOMASK), me.name, parv[0], sptr->snomask,
          sptr->snomask);
    return 0;
  }

  /*
   * find flags already set for user
   * why not just copy them?
   */
  setflags = sptr->flags;

  if (MyConnect(sptr))
    tmpmask = sptr->snomask;

  /*
   * parse mode change string(s)
   */
  for (p = &parv[2]; *p; p++) {       /* p is changed in loop too */
    for (m = *p; *m; m++) {
      switch (*m) {
      case '+':
        what = MODE_ADD;
        break;
      case '-':
        what = MODE_DEL;
        break;
      case 's':
        if (*(p + 1) && is_snomask(*(p + 1))) {
          snomask_given = 1;
          tmpmask = umode_make_snomask(tmpmask, *++p, what);
          tmpmask &= (IsAnOper(sptr) ? SNO_ALL : SNO_USER);
        }
        else
          tmpmask = (what == MODE_ADD) ?
              (IsAnOper(sptr) ? SNO_OPERDEFAULT : SNO_DEFAULT) : 0;
        if (tmpmask)
          SetServNotice(sptr);
        else
          ClearServNotice(sptr);
        break;
      case 'w':
        if (what == MODE_ADD)
          SetWallops(sptr);
        else
          ClearWallops(sptr);
        break;
      case 'o':
        if (what == MODE_ADD)
          SetOper(sptr);
        else {
          sptr->flags &= ~(FLAGS_OPER | FLAGS_LOCOP);
          if (MyConnect(sptr)) {
            tmpmask = sptr->snomask & ~SNO_OPER;
            sptr->handler = CLIENT_HANDLER;
          }
        }
        break;
      case 'O':
        if (what == MODE_ADD)
          SetLocOp(sptr);
        else { 
          sptr->flags &= ~(FLAGS_OPER | FLAGS_LOCOP);
          if (MyConnect(sptr)) {
            tmpmask = sptr->snomask & ~SNO_OPER;
            sptr->handler = CLIENT_HANDLER;
          }
        }
        break;
      case 'i':
        if (what == MODE_ADD)
          SetInvisible(sptr);
        else
          ClearInvisible(sptr);
        break;
      case 'd':
        if (what == MODE_ADD)
          SetDeaf(sptr);
        else
          ClearDeaf(sptr);
        break;
      case 'k':
        if (what == MODE_ADD)
          SetChannelService(sptr);
        else
          ClearChannelService(sptr);
        break;
      case 'L':
        if (what == MODE_ADD)
          SetLib(sptr);
        else
          ClearLib(sptr);
        break;
      case 'p':
        if (what == MODE_ADD)
         {
          SetHide(sptr);
         if (MyUser(sptr))
          strncpy(sptr->sockhost,sptr->user->crypthost,HOSTLEN);
         if (sptr->user) strncpy(sptr->user->host, sptr->user->crypthost, HOSTLEN);
         }
        else
        {
          ClearHide(sptr);
         if (MyUser(sptr))
          strncpy(sptr->sockhost,sptr->user->orighost,HOSTLEN);
         if (sptr->user) strncpy(sptr->user->host, sptr->user->orighost, HOSTLEN);
         }
        break;
      case 'g':
        if (what == MODE_ADD)
          SetDebug(sptr);
        else
          ClearDebug(sptr);
        break;
      default:
        break;
      }
    }
  }
  /*
   * Evaluate rules for new user mode
   * Stop users making themselves operators too easily:
   */
  if (!(setflags & FLAGS_OPER) && IsOper(sptr) && !IsServer(cptr))
    ClearOper(sptr);
  if (!(setflags & FLAGS_LOCOP) && IsLocOp(sptr) && !IsServer(cptr))
    ClearLocOp(sptr);
#ifdef WALLOPS_OPER_ONLY
  /*
   * only send wallops to opers
   */
  if (!IsAnOper(sptr) && !(setflags & FLAGS_WALLOP) && !IsServer(cptr))
    ClearWallops(sptr);
#endif
#ifdef SERVNOTICE_OPER_ONLY
  if (MyConnect(sptr) && !IsAnOper(sptr) && !(setflags & FLAGS_SERVNOTICE)) {
    ClearServNotice(sptr);
    sptr->snomask = 0;
  }
#endif
#ifdef DEBUG_OPER_ONLY
  if (!IsAnOper(sptr) && !(setflags & FLAGS_DEBUG))
    ClearDebug(sptr);
#endif
  if (MyConnect(sptr)) {
    if ((setflags & (FLAGS_OPER | FLAGS_LOCOP)) && !IsAnOper(sptr))
      det_confs_butmask(sptr, CONF_CLIENT & ~CONF_OPS);

    if (SendServNotice(sptr)) {
      if (tmpmask != sptr->snomask)
	set_snomask(sptr, tmpmask, SNO_SET);
      if (sptr->snomask && snomask_given)
	sendto_one(sptr, rpl_str(RPL_SNOMASK), me.name, parv[0],
		   sptr->snomask, sptr->snomask);
    } else
      set_snomask(sptr, 0, SNO_SET);
  }
  /*
   * new umode; servers can set it, local users cannot;
   * prevents users from /kick'ing or /mode -o'ing
   */
  if ((setflags & FLAGS_LIBERIX) && !IsOper(sptr) && !IsServer(cptr))
    ClearLib(sptr);
  if (!(setflags & FLAGS_CHSERV) && !IsServer(cptr))
    ClearChannelService(sptr);
  if (!(setflags & FLAGS_LIBERIX) && !IsOper(sptr))
    ClearLib(sptr);
  /*
   * Compare new flags with old flags and send string which
   * will cause servers to update correctly.
   */
  if ((setflags & FLAGS_OPER) && !IsOper(sptr))
    --UserStats.opers;
  if (!(setflags & FLAGS_OPER) && IsOper(sptr))
    ++UserStats.opers;
  if ((setflags & FLAGS_INVISIBLE) && !IsInvisible(sptr))
    --UserStats.inv_clients;
  if (!(setflags & FLAGS_INVISIBLE) && IsInvisible(sptr))
    ++UserStats.inv_clients;
  send_umode_out(cptr, sptr, setflags);

#ifdef SERVNOTICE_OPER_ONLY
  if (MyConnect(sptr) && IsAnOper(sptr)) {
#else
  if (MyConnect(sptr)) {
#endif
    if (tmpmask != sptr->snomask)
      set_snomask(sptr, tmpmask, SNO_SET);
    if (sptr->snomask && snomask_given)
      sendto_one(sptr, rpl_str(RPL_SNOMASK), me.name, sptr->name,
                 sptr->snomask, sptr->snomask);
  }

  return 0;
}

/*
 * Build umode string for BURST command
 * --Run
 */
char *umode_str(struct Client *cptr)
{
  char* m = umodeBuf;                /* Maximum string size: "owidg\0" */
  int   i;
  int   c_flags;

  c_flags = cptr->flags & SEND_UMODES;        /* cleaning up the original code */

  for (i = 0; i < USERMODELIST_SIZE; ++i) {
    if ( (c_flags & userModeList[i].flag))
      *m++ = userModeList[i].c;
  }
  if (IsAccount(cptr)) {
    char *t = cptr->user->account;

    *m++ = ' ';
    while ((*m++ = *t++))
      ; /* Empty loop */
  }
  *m = '\0';

  return umodeBuf;                /* Note: static buffer, gets
                                   overwritten by send_umode() */
}

/*
 * Send the MODE string for user (user) to connection cptr
 * -avalon
 */
void send_umode(struct Client *cptr, struct Client *sptr, int old, int sendmask)
{
  int i;
  int flag;
  char *m;
  int what = MODE_NULL;

  /*
   * Build a string in umodeBuf to represent the change in the user's
   * mode between the new (sptr->flag) and 'old'.
   */
  m = umodeBuf;
  *m = '\0';
  for (i = 0; i < USERMODELIST_SIZE; ++i) {
    flag = userModeList[i].flag;
    if (MyUser(sptr) && !(flag & sendmask))
      continue;
    if ( (flag & old) && !(sptr->flags & flag))
    {
      if (what == MODE_DEL)
        *m++ = userModeList[i].c;
      else
      {
        what = MODE_DEL;
        *m++ = '-';
        *m++ = userModeList[i].c;
      }
    }
    else if (!(flag & old) && (sptr->flags & flag))
    {
      if (what == MODE_ADD)
        *m++ = userModeList[i].c;
      else
      {
        what = MODE_ADD;
        *m++ = '+';
        *m++ = userModeList[i].c;
      }
    }
  }
  *m = '\0';
  if (*umodeBuf && cptr)
    sendto_one(cptr, ":%s MODE %s :%s", sptr->name, sptr->name, umodeBuf);
}

/*
 * Check to see if this resembles a sno_mask.  It is if 1) there is
 * at least one digit and 2) The first digit occurs before the first
 * alphabetic character.
 */
int is_snomask(char *word)
{
  if (word)
  {
    for (; *word; word++)
      if (IsDigit(*word))
        return 1;
      else if (IsAlpha(*word))
        return 0;
  }
  return 0;
}

/*
 * If it begins with a +, count this as an additive mask instead of just
 * a replacement.  If what == MODE_DEL, "+" has no special effect.
 */
unsigned int umode_make_snomask(unsigned int oldmask, char *arg, int what)
{
  unsigned int sno_what;
  unsigned int newmask;
  if (*arg == '+')
  {
    arg++;
    if (what == MODE_ADD)
      sno_what = SNO_ADD;
    else
      sno_what = SNO_DEL;
  }
  else if (*arg == '-')
  {
    arg++;
    if (what == MODE_ADD)
      sno_what = SNO_DEL;
    else
      sno_what = SNO_ADD;
  }
  else
    sno_what = (what == MODE_ADD) ? SNO_SET : SNO_DEL;
  /* pity we don't have strtoul everywhere */
  newmask = (unsigned int)atoi(arg);
  if (sno_what == SNO_DEL)
    newmask = oldmask & ~newmask;
  else if (sno_what == SNO_ADD)
    newmask |= oldmask;
  return newmask;
}

static void delfrom_list(struct Client *cptr, struct SLink **list)
{
  struct SLink* tmp;
  struct SLink* prv = NULL;

  for (tmp = *list; tmp; tmp = tmp->next) {
    if (tmp->value.cptr == cptr) {
      if (prv)
        prv->next = tmp->next;
      else
        *list = tmp->next;
      free_link(tmp);
      break;
    }
    prv = tmp;
  }
}

/*
 * This function sets a Client's server notices mask, according to
 * the parameter 'what'.  This could be even faster, but the code
 * gets mighty hard to read :)
 */
void set_snomask(struct Client *cptr, unsigned int newmask, int what)
{
  unsigned int oldmask, diffmask;        /* unsigned please */
  int i;
  struct SLink *tmp;

  oldmask = cptr->snomask;

  if (what == SNO_ADD)
    newmask |= oldmask;
  else if (what == SNO_DEL)
    newmask = oldmask & ~newmask;
  else if (what != SNO_SET)        /* absolute set, no math needed */
    sendto_ops("setsnomask called with %d ?!", what);

  newmask &= (IsAnOper(cptr) ? SNO_ALL : SNO_USER);

  diffmask = oldmask ^ newmask;

  for (i = 0; diffmask >> i; i++) {
    if (((diffmask >> i) & 1))
    {
      if (((newmask >> i) & 1))
      {
        tmp = make_link();
        tmp->next = opsarray[i];
        tmp->value.cptr = cptr;
        opsarray[i] = tmp;
      }
      else
        /* not real portable :( */
        delfrom_list(cptr, &opsarray[i]);
    }
  }
  cptr->snomask = newmask;
}

/*
 * is_silenced : Does the actual check wether sptr is allowed
 *               to send a message to acptr.
 *               Both must be registered persons.
 * If sptr is silenced by acptr, his message should not be propagated,
 * but more over, if this is detected on a server not local to sptr
 * the SILENCE mask is sent upstream.
 */
int is_silenced(struct Client *sptr, struct Client *acptr)
{
  struct SLink *lp;
  struct User *user;
  static char sender[HOSTLEN + NICKLEN + USERLEN + 5];
  static char senderip[16 + NICKLEN + USERLEN + 5];

  if (!(acptr->user) || !(lp = acptr->user->silence) || !(user = sptr->user))
    return 0;
  sprintf_irc(sender, "%s!%s@%s", sptr->name, user->username, user->host);
  sprintf_irc(senderip, "%s!%s@%s", sptr->name, user->username,
              ircd_ntoa((const char*) &sptr->ip));
  for (; lp; lp = lp->next)
  {
    if ((!(lp->flags & CHFL_SILENCE_IPMASK) && !match(lp->value.cp, sender)) ||
        ((lp->flags & CHFL_SILENCE_IPMASK) && !match(lp->value.cp, senderip)))
    {
      if (!MyConnect(sptr))
      {
        if (Protocol(sptr->from) < 10) {
	  if (IsUser(acptr))
	    sendto_one(sptr->from, "%s%s " TOK_SILENCE " %s %s",
		       NumNick(acptr), NumServ(sptr), lp->value.cp);
	  else
	    sendto_one(sptr->from, "%s " TOK_SILENCE " %s %s", NumServ(acptr),
		       NumServ(sptr), lp->value.cp);
	} else {
	  if (IsUser(acptr))
	    sendto_one(sptr->from, "%s%s " TOK_SILENCE " %s%s %s",
		       NumNick(acptr), NumNick(sptr), lp->value.cp);
	  else
	    sendto_one(sptr->from, "%s " TOK_SILENCE " %s%s %s",
		       NumServ(acptr), NumNick(sptr), lp->value.cp);
	}
      }
      return 1;
    }
  }
  return 0;
}

/*
 * del_silence
 *
 * Removes all silence masks from the list of sptr that fall within `mask'
 * Returns -1 if none where found, 0 otherwise.
 */
int del_silence(struct Client *sptr, char *mask)
{
  struct SLink **lp;
  struct SLink *tmp;
  int ret = -1;

  for (lp = &sptr->user->silence; *lp;) {
    if (!mmatch(mask, (*lp)->value.cp))
    {
      tmp = *lp;
      *lp = tmp->next;
      MyFree(tmp->value.cp);
      free_link(tmp);
      ret = 0;
    }
    else
      lp = &(*lp)->next;
  }
  return ret;
}

int add_silence(struct Client* sptr, const char* mask)
{
  struct SLink *lp, **lpp;
  int cnt = 0, len = strlen(mask);
  char *ip_start;

  for (lpp = &sptr->user->silence, lp = *lpp; lp;)
  {
    if (0 == ircd_strcmp(mask, lp->value.cp))
      return -1;
    if (!mmatch(mask, lp->value.cp))
    {
      struct SLink *tmp = lp;
      *lpp = lp = lp->next;
      MyFree(tmp->value.cp);
      free_link(tmp);
      continue;
    }
    if (MyUser(sptr))
    {
      len += strlen(lp->value.cp);
      if ((len > MAXSILELENGTH) || (++cnt >= MAXSILES))
      {
        sendto_one(sptr, err_str(ERR_SILELISTFULL), me.name, sptr->name, mask);
        return -1;
      }
      else if (!mmatch(lp->value.cp, mask))
        return -1;
    }
    lpp = &lp->next;
    lp = *lpp;
  }
  lp = make_link();
  memset(lp, 0, sizeof(struct SLink));
  lp->next = sptr->user->silence;
  lp->value.cp = (char*) MyMalloc(strlen(mask) + 1);
  assert(0 != lp->value.cp);
  strcpy(lp->value.cp, mask);
  if ((ip_start = strrchr(mask, '@')) && check_if_ipmask(ip_start + 1))
    lp->flags = CHFL_SILENCE_IPMASK;
  sptr->user->silence = lp;
  return 0;
}

int
send_supported(struct Client *cptr)
{
  char             featurebuf[512];

  sprintf_irc(featurebuf, FEATURES1, FEATURESVALUES1);
  sendto_one(cptr, rpl_str(RPL_ISUPPORT), me.name, cptr->name, featurebuf);
  sprintf_irc(featurebuf, FEATURES2, FEATURESVALUES2);
  sendto_one(cptr, rpl_str(RPL_ISUPPORT), me.name, cptr->name, featurebuf);

  return 0; /* convenience return, if it's ever needed */
}
int ms_account(struct Client* cptr, struct Client* sptr, int parc,
	       char* parv[])
{
  struct Client *acptr;
  if (parc < 3 || !IsServer(sptr)) 
    /* Message is invalid, ignore it */
    return 0;

  if (!(acptr = findNUser(parv[1])))
    return 0; /* Ignore ACCOUNT for a user that QUIT; probably crossed */

  if (IsAccount(acptr))
    /* Can't auth twice */
    return 0;

  assert(0 == acptr->user->account[0]);

  ircd_strncpy(acptr->user->account, parv[2], ACCOUNTLEN);
  SetAccount(acptr);

  sendto_serv_butone(sptr, "%s " TOK_ACCOUNT " %s %s", NumServ(sptr), parv[1], acptr->user->account);

  return 0;
}
